use std::{
    collections::VecDeque,
    ops::{AddAssign, Div, Sub},
};

const MODE_M_MAX: usize = 4;
const M: i32 = 2;
const N_SYM: i32 = 35;

#[link(name = "fsk", kind = "static")]
extern "C" {
    fn fsk_create_hbr(
        Fs: libc::c_int,
        Rs: libc::c_int,
        M: libc::c_int,
        P: libc::c_int,
        n_sym: libc::c_int,
        f1_tx: libc::c_int,
        tone_spacing: libc::c_int,
    ) -> *mut InternalFsk;

    fn fsk_destroy(fsk: *mut InternalFsk);

    fn fsk_demod_sd(fsk: *mut InternalFsk, rx_filt: *mut libc::c_float, fsk_in: *const Complex);

    fn fsk_rx_filt_to_llrs(
        llr: *mut libc::c_float,
        rx_filt: *const libc::c_float,
        v_est: libc::c_float,
        snr_est: libc::c_float,
        m: libc::c_int,
        n_syms: libc::c_int,
    );

    fn fsk_set_freq_est_limits(fsk: *mut InternalFsk, f_min: libc::c_int, f_max: libc::c_int);
}

#[repr(C)]
struct InternalFsk {
    /*  Static parameters set up by fsk_init */
    n_dft: libc::c_int, /* freq offset est fft */
    fs: libc::c_int,    /* sample freq */
    n: libc::c_int,     /* processing buffer size */
    rs: libc::c_int,    /* symbol rate */
    ts: libc::c_int,    /* samples per symbol */
    n_mem: libc::c_int, /* size of extra mem for timing adj */
    p: libc::c_int,     /* oversample rate for timing est/adj */
    n_sym: libc::c_int, /* Number of symbols processed by demodulator in each call, also the
                        timing estimator window */
    n_bits: libc::c_int, /* Number of bits spat out in a processing frame */
    f1_tx: libc::c_int,  /* f1 for modulator */
    tone_spacing: libc::c_int, /* Space between TX freqs for modulator (and option mask
                         freq estimator) */
    mode: libc::c_int,              /* 2FSK or 4FSK */
    tc: libc::c_float,              /* time constant for smoothing FFTs */
    est_min: libc::c_int,           /* Minimum frequency for freq. estimator */
    est_max: libc::c_int,           /* Maximum frequency for freq. estimator */
    est_space: libc::c_int,         /* Minimum frequency spacing for freq. estimator */
    hann_table: *mut libc::c_float, /* Precomputed or runtime computed hann window table */

    /*  Parameters used by demod */
    s_f: *mut libc::c_float,      /* Average of magnitude spectrum */
    phi_c: [Complex; MODE_M_MAX], /* phase of each demod local oscillator */
    f_dc: *mut Complex,           /* down converted samples               */

    fft_cfg: *mut libc::c_void, /* Config for KISS FFT, used in freq est */
    norm_rx_timing: libc::c_float, /* Normalized RX timing */

    /*  Parameters used by mod */
    tx_phase_c: Complex, /* TX phase, but complex */

    /*  Statistics generated by demod */
    eb_no_db: libc::c_float,             /* Estimated EbNo in dB */
    f_est: [libc::c_float; MODE_M_MAX],  /* Estimated frequencies (peak method) */
    f2_est: [libc::c_float; MODE_M_MAX], /* Estimated frequencies (mask method) */
    freq_est_type: libc::c_int,          /* which estimator to use              */
    ppm: libc::c_float,                  /* Estimated PPM clock offset */
    snr_est: libc::c_float,              /* used for LLRs */
    v_est: libc::c_float,                /* used for LLRs */
    rx_sig_pow: libc::c_float,
    rx_nse_pow: libc::c_float,

    /*  Parameters used by mod/demod and driving code */
    nin: libc::c_int,        /* Number of samples to feed the next demod cycle */
    burst_mode: libc::c_int, /* enables/disables 'burst' mode */
    lock_nin: libc::c_int,   /* locks nin during testing */

    /*  modem statistics struct */
    stats: *mut libc::c_void,
    normalise_eye: libc::c_int, /* enables/disables normalisation of eye diagram */
}

#[repr(C)]
#[derive(Debug, Copy, Clone)]
pub struct Complex {
    pub real: libc::c_float,
    pub imag: libc::c_float,
}

impl Complex {
    pub const fn zero() -> Self {
        Self {
            real: 0.0,
            imag: 0.0,
        }
    }
}

impl Sub for Complex {
    type Output = Complex;

    fn sub(self, rhs: Self) -> Complex {
        Complex {
            real: self.real - rhs.real,
            imag: self.imag - rhs.imag,
        }
    }
}

impl AddAssign for Complex {
    fn add_assign(&mut self, rhs: Self) {
        self.real += rhs.real;
        self.imag += rhs.imag;
    }
}

impl Div<f32> for Complex {
    type Output = Complex;

    fn div(self, rhs: f32) -> Complex {
        Complex {
            real: self.real / rhs,
            imag: self.imag / rhs,
        }
    }
}

pub struct Fsk<I: Iterator<Item = Complex>> {
    internal: *mut InternalFsk,
    input_cache: Vec<Complex>,
    output_cache: VecDeque<f32>,
    iq_iter: I,
}

impl<I: Iterator<Item = Complex>> Fsk<I> {
    pub fn new(iq_iter: I) -> Self {
        const P: i32 = 5;
        const FS: i32 = 9600 * P;

        let internal = unsafe { fsk_create_hbr(FS, 9600, M, P, N_SYM, -1, 0) };

        // Set upper/lower bound on the FSK peaks, from center
        // (Hz)
        let fsk_lower = -10_000;
        let fsk_upper = 10_000;

        unsafe { fsk_set_freq_est_limits(internal, fsk_lower, fsk_upper) };

        Self {
            internal,
            iq_iter,
            input_cache: vec![],
            output_cache: VecDeque::new(),
        }
    }
}

impl<I: Iterator<Item = Complex>> Iterator for Fsk<I> {
    type Item = f32;

    fn next(&mut self) -> Option<Self::Item> {
        if let Some(x) = self.output_cache.pop_front() {
            return Some(x);
        }

        let nin = unsafe { (*self.internal).nin }.try_into().unwrap();
        while self.input_cache.len() < nin {
            self.input_cache.push(self.iq_iter.next()?);
        }

        let n_bits = unsafe { (*self.internal).n_bits };
        self.output_cache.resize(n_bits.try_into().unwrap(), 0.0);

        let n_sym = unsafe { (*self.internal).n_sym };
        let mut rx_filt = vec![0.0; usize::try_from(M * n_sym).unwrap()];
        let rx_filt = rx_filt.as_mut_ptr();
        let fsk_in = self.input_cache.as_ptr();
        let llrs = self.output_cache.make_contiguous().as_mut_ptr();
        unsafe {
            fsk_demod_sd(self.internal, rx_filt, fsk_in);

            fsk_rx_filt_to_llrs(
                llrs,
                rx_filt,
                (*self.internal).v_est,
                (*self.internal).snr_est,
                (*self.internal).mode,
                (*self.internal).n_sym,
            );
        }
        self.input_cache.clear();

        Some(self.output_cache.pop_front().unwrap())
    }
}

impl<I: Iterator<Item = Complex>> Drop for Fsk<I> {
    fn drop(&mut self) {
        unsafe {
            fsk_destroy(self.internal);
        }
    }
}
